@@ -1,14 +1,14 @@
-use once_cell::sync::OnceCell;
 use azure_core::{prelude::IfMatchCondition, HttpClient};
+pub use azure_cosmos::prelude::PartitionKeys;
 use azure_cosmos::prelude::*;
 use azure_storage::clients::*;
 use futures::StreamExt;
+use once_cell::sync::OnceCell;
 use serde::{de::DeserializeOwned, Serialize};
 use std::sync::Arc;
 use tokio::time::{sleep, Duration};
 use uuid::Uuid;
 use warp::{filters::multipart::FormData, Buf as OtherBuf};
-pub use azure_cosmos::prelude::PartitionKeys;
 
 /// This struct keeps the static state of cosmos DB, it must be initialized before any cosmos
 /// functions are called.
@@ -87,25 +87,28 @@ impl warp::reject::Reject for CosmosErrorStruct {}
 
 /// Utility function that returns a closure that converts whatever error into a Reject error.
 /// Usage is: function_call_which_returns_non_reject_result().map_err(into_cosmos_error("custom error message"))?;
-fn into_cosmos_error<E: ToString, S: ToString>(message: S) -> impl FnOnce(E) -> CosmosError {
-    move |e: E| new_cosmos_error(format!("{} : {:?}", message.to_string(), e.to_string()))
+fn into_cosmos_error<M: ToString>(msg: M) -> impl FnOnce(Box<dyn std::error::Error + Sync + Send>) -> CosmosError {
+    move |e: Box<dyn std::error::Error + Sync + Send>| new_cosmos_error(msg.to_string(), e)
 }
 
 /// Creates a new cosmos error from a given error
-fn new_cosmos_error<E: ToString>(err: E) -> CosmosError {
-    let err = err.to_string();
-    let kind = if err.contains("412") {
-        CosmosErrorKind::PreconditionFailed
-    } else if err.contains("404") {
-        CosmosErrorKind::NotFound
-    } else if err.contains("400") {
-        CosmosErrorKind::BadRequest
-    } else if err.contains("409") {
-        CosmosErrorKind::Conflict
-    } else {
-        CosmosErrorKind::InternalError
+fn new_cosmos_error<M: ToString>(msg: M, err: Box<dyn std::error::Error>) -> CosmosError {
+    let kind = match err.downcast_ref::<azure_core::errors::UnexpectedHTTPResult>() {
+        Some(err) => match err.status_code().as_u16() {
+            412 => CosmosErrorKind::PreconditionFailed,
+            404 => CosmosErrorKind::NotFound,
+            400 => CosmosErrorKind::BadRequest,
+            409 => CosmosErrorKind::Conflict,
+            _ => CosmosErrorKind::InternalError,
+        },
+        None => CosmosErrorKind::InternalError,
     };
-    CosmosErrorStruct { kind, err }
+    let msg = format!("{} : {}", msg.to_string(), err.to_string());
+    CosmosErrorStruct { kind, err: msg }
+}
+
+fn new_cosmos_error_internal<M: ToString>(msg: M) -> CosmosError {
+    CosmosErrorStruct { kind: CosmosErrorKind::InternalError, err: msg.to_string() }
 }
 
 /// Creates a new cosmos error from a given error and the error kind
@@ -121,16 +124,20 @@ async fn insert_internal<D: Serialize, P: Into<PartitionKeys>, C: ToString>(
     etag: Option<&str>,
     upsert: bool,
 ) -> Result<String, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token"))?;
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name);
 
@@ -147,10 +154,10 @@ async fn insert_internal<D: Serialize, P: Into<PartitionKeys>, C: ToString>(
         match c.execute(&Document::new(document)).await {
             Ok(t) => Ok(t),
             Err(err) => {
-                return Err(RetryLoopError::Permanent(new_cosmos_error(format!(
-                    "Cosmos db error: {:?}",
-                    err
-                ))));
+                return Err(RetryLoopError::Permanent(new_cosmos_error(
+                    format!("Cosmos db error"),
+                    err,
+                )));
             }
         }
     })
@@ -187,17 +194,21 @@ pub async fn get<D: DeserializeOwned, P: Into<PartitionKeys>, C: ToString, S: To
     pk: P,
     document_id: S,
 ) -> Result<(D, String), CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token"))?;
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -240,17 +251,21 @@ pub async fn modify_async_get_old<
     document_id: S,
     transform: F,
 ) -> Result<(D, D, String), CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token"))?;
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -262,10 +277,10 @@ pub async fn modify_async_get_old<
             .execute::<D>()
             .await
             .map_err(|e| {
-                RetryLoopError::Permanent(new_cosmos_error(format!(
-                    "Could not get document {:?}",
-                    e
-                )))
+                RetryLoopError::Permanent(new_cosmos_error(
+                    format!("Could not get document"),
+                    e,
+                ))
             })? {
             azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
             azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
@@ -296,20 +311,29 @@ pub async fn modify_async_get_old<
                 return Result::Ok::<_, RetryLoopError<CosmosError>>((doc, old_doc, resp.etag))
             }
             Err(err) => {
-                let s = err.to_string();
+                let err = new_cosmos_error(format!("Cosmos db error"), err);
                 //NOTE: 412 means the document has been edited between read and write so it
                 //means we need to retry the entire read/write block
-                if s.contains("412") {
-                    return Err(RetryLoopError::Transient(new_cosmos_error_kind(
-                        format!("Cosmos db error: {:?}", s),
-                        CosmosErrorKind::PreconditionFailed,
-                    )));
-                } else {
-                    return Err(RetryLoopError::Permanent(new_cosmos_error(format!(
-                        "Cosmos db error: {:?}",
-                        s
-                    ))));
+                match err.kind {
+                    CosmosErrorKind::PreconditionFailed => {
+                        return Err(RetryLoopError::Transient(err));
+                    },
+                    _ => {
+                        return Err(RetryLoopError::Permanent(err));
+                    }
                 }
+                //let s = err.to_string();
+                //if s.contains("412") {
+                //    return Err(RetryLoopError::Transient(new_cosmos_error_kind(
+                //        format!("Cosmos db error: {:?}", s),
+                //        CosmosErrorKind::PreconditionFailed,
+                //    )));
+                //} else {
+                //    return Err(RetryLoopError::Permanent(new_cosmos_error(
+                //        format!("Cosmos db error"),
+                //        err,
+                //    )));
+                //}
             }
         };
     })
@@ -334,17 +358,21 @@ pub async fn modify_async<
     document_id: S,
     transform: F,
 ) -> Result<D, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token"))?;
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -356,10 +384,10 @@ pub async fn modify_async<
             .execute::<D>()
             .await
             .map_err(|e| {
-                RetryLoopError::Permanent(new_cosmos_error(format!(
-                    "Could not get document {:?}",
-                    e
-                )))
+                RetryLoopError::Permanent(new_cosmos_error(
+                    format!("Could not get document"),
+                    e,
+                ))
             })? {
             azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
             azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
@@ -387,19 +415,16 @@ pub async fn modify_async<
         match c.execute(&Document::new(&doc)).await {
             Ok(_) => return Result::Ok::<_, RetryLoopError<CosmosError>>(doc),
             Err(err) => {
-                let s = err.to_string();
+                let err = new_cosmos_error(format!("Cosmos db error"), err);
                 //NOTE: 412 means the document has been edited between read and write so it
                 //means we need to retry the entire read/write block
-                if s.contains("412") {
-                    return Err(RetryLoopError::Transient(new_cosmos_error_kind(
-                        format!("Cosmos db error: {:?}", s),
-                        CosmosErrorKind::PreconditionFailed,
-                    )));
-                } else {
-                    return Err(RetryLoopError::Permanent(new_cosmos_error(format!(
-                        "Cosmos db error: {:?}",
-                        s
-                    ))));
+                match err.kind {
+                    CosmosErrorKind::PreconditionFailed => {
+                        return Err(RetryLoopError::Transient(err));
+                    },
+                    _ => {
+                        return Err(RetryLoopError::Permanent(err));
+                    }
                 }
             }
         };
@@ -438,17 +463,21 @@ pub async fn maybe_modify<
     document_id: S,
     transform: F,
 ) -> Result<ModifyReturn<D>, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token"))?;
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -459,7 +488,7 @@ pub async fn maybe_modify<
             .get_document()
             .execute::<D>()
             .await
-            .map_err(|e| RetryLoopError::Permanent(new_cosmos_error(format!("{:?}", e))))?
+            .map_err(|e| RetryLoopError::Permanent(new_cosmos_error(format!(""), e)))?
         {
             azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
             azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
@@ -495,19 +524,16 @@ pub async fn maybe_modify<
                 return Result::Ok::<_, RetryLoopError<CosmosError>>(ModifyReturn::Replace(doc))
             }
             Err(err) => {
-                let s = err.to_string();
+                let err = new_cosmos_error(format!("Cosmos db error"), err);
                 //NOTE: 412 means the document has been edited between read and write so it
                 //means we need to retry the entire read/write block
-                if s.contains("412") {
-                    return Err(RetryLoopError::Transient(new_cosmos_error_kind(
-                        format!("Cosmos db error: {:?}", s),
-                        CosmosErrorKind::PreconditionFailed,
-                    )));
-                } else {
-                    return Err(RetryLoopError::Transient(new_cosmos_error(format!(
-                        "Cosmos db error: {:?}",
-                        s
-                    ))));
+                match err.kind {
+                    CosmosErrorKind::PreconditionFailed => {
+                        return Err(RetryLoopError::Transient(err));
+                    },
+                    _ => {
+                        return Err(RetryLoopError::Permanent(err));
+                    }
                 }
             }
         };
@@ -530,17 +556,21 @@ pub async fn modify_no_retry<
     document_id: S,
     transform: F,
 ) -> Result<D, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token"))?;
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -550,7 +580,7 @@ pub async fn modify_no_retry<
         .get_document()
         .execute::<D>()
         .await
-        .map_err(|e| new_cosmos_error(format!("{:?}", e)))?
+        .map_err(|e| new_cosmos_error(format!(""), e))?
     {
         azure_cosmos::responses::GetDocumentResponse::Found(resp) => resp,
         azure_cosmos::responses::GetDocumentResponse::NotFound(resp) => {
@@ -578,7 +608,7 @@ pub async fn modify_no_retry<
     let doc = match c.execute(&Document::new(&doc)).await {
         Ok(_) => doc,
         Err(err) => {
-            return Err(new_cosmos_error(format!("Cosmos db error: {:?}", err)));
+            return Err(new_cosmos_error(format!("Cosmos db error"), err));
         }
     };
     Ok(doc)
@@ -609,17 +639,21 @@ pub async fn delete<C: ToString, S: ToString, P: Into<PartitionKeys>>(
     document_id: S,
     etag: Option<String>,
 ) -> Result<(), CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let document_id = document_id.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token"))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token"))?;
 
     let pk = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name);
     let document_client = collection_client
@@ -648,15 +682,19 @@ pub async fn query_crosspartition_etag<D: DeserializeOwned, P: Into<PartitionKey
     max_count: i32,
     cross_partition: bool,
 ) -> Result<Vec<(D, String)>, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     let collection_name = collection_name.to_string();
     let authorization_token = AuthorizationToken::primary_from_base64(&state.cosmos_master_key)
-        .map_err(into_cosmos_error("Could not get authorization token."))?;
+        .map_err(|_| new_cosmos_error_internal("Could not get authorization token."))?;
     let pk: PartitionKeys = pk.into();
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
-    let client = CosmosClient::new(http_client, state.cosmos_account.to_string(), authorization_token);
+    let client = CosmosClient::new(
+        http_client,
+        state.cosmos_account.to_string(),
+        authorization_token,
+    );
     let database_client = client.into_database_client(&state.cosmos_database);
     let collection_client = database_client.into_collection_client(collection_name.clone());
     let mut documents: Vec<(D, String)> = vec![];
@@ -686,7 +724,7 @@ pub async fn query_crosspartition_etag<D: DeserializeOwned, P: Into<PartitionKey
         let query_documents_response =
             query_documents_response
                 .into_documents()
-                .map_err(into_cosmos_error(
+                .map_err(|_| new_cosmos_error_internal(
                     "Could not get cosmos db query document response.",
                 ))?;
         let mut fetched_documents: Vec<(D, String)> = query_documents_response
@@ -732,12 +770,17 @@ pub async fn query<D: DeserializeOwned, P: Into<PartitionKeys>, C: ToString>(
 }
 
 /// Gets the data from a blob storage
-pub async fn get_blob<S: AsRef<str>>(id: S, storage_container: &str) -> Result<Vec<u8>, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+pub async fn get_blob<S: AsRef<str>>(
+    id: S,
+    storage_container: &str,
+) -> Result<Vec<u8>, CosmosError> {
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     if state.storage_account.is_none() || state.storage_master_key.is_none() {
-        return Err(new_cosmos_error(format!("Cosmos storage values are not initialized")));
+        return Err(new_cosmos_error_internal(format!(
+            "Cosmos storage values are not initialized"
+        )));
     }
     let http_client: Arc<Box<dyn HttpClient>> = Arc::new(Box::new(reqwest::Client::new()));
     let blob_client = StorageAccountClient::new_access_key(
@@ -767,11 +810,13 @@ pub async fn upload_blob(
     expected_content_type: &str,
     storage_container: &str,
 ) -> Result<String, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     if state.storage_account.is_none() || state.storage_master_key.is_none() {
-        return Err(new_cosmos_error(format!("Cosmos storage values are not initialized")));
+        return Err(new_cosmos_error_internal(format!(
+            "Cosmos storage values are not initialized"
+        )));
     }
     while let Some(r) = f.next().await {
         match r {
@@ -890,13 +935,21 @@ pub async fn upload_blob(
 
 /// Uploads a new form data image to the blob storage and returns the image_id
 pub async fn upload_image(f: FormData) -> Result<String, CosmosError> {
-    let state = COSMOS_STATE.get().ok_or_else(|| {
-        new_cosmos_error(format!("Cosmos state not initialized"))
-    })?;
+    let state = COSMOS_STATE
+        .get()
+        .ok_or_else(|| new_cosmos_error_internal(format!("Cosmos state not initialized")))?;
     if state.image_storage_container.is_none() {
-        return Err(new_cosmos_error(format!("Cosmos image storage container is not initialized")));
+        return Err(new_cosmos_error_internal(format!(
+            "Cosmos image storage container is not initialized"
+        )));
     }
-    upload_blob(f, "image", "image", &state.image_storage_container.as_ref().unwrap()).await
+    upload_blob(
+        f,
+        "image",
+        "image",
+        &state.image_storage_container.as_ref().unwrap(),
+    )
+    .await
 }
 
 pub struct CosmosSaga {
